/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 */

#include "ldap-int.h"
#include <gssapi/gssapi_generic.h>

#define NSOCKD 10
struct {
    int sd;
    gss_ctx_id_t gss_context;
    char *buf, *p;
    int len;
} sockd_list[NSOCKD];

size_t
gss_read(int sd, char *buf, size_t len);

size_t
gss_write(int sd, char *buf, size_t len);

void
gss_display_error( char *where, OM_uint32 maj, OM_uint32 min ) {
    OM_uint32 maj_stat, min_stat;
    gss_buffer_desc status_buffer;
    int ctx = 0;
 
    LDAPDebug(LDAP_DEBUG_ANY, "GSS Auth Errors detected in %s\n", where, 0, 0);
    do {
   
        if (GSS_ERROR(gss_display_status(&min_stat,
                                     maj,
                                     GSS_C_GSS_CODE,
                                     GSS_C_NULL_OID,
                                     &ctx,
                                     &status_buffer))) {
          LDAPDebug(LDAP_DEBUG_ANY, "Error translating major status code (%X)\n",
                   maj, 0, 0);
          ctx = 0;
      } else {
          LDAPDebug(LDAP_DEBUG_ANY, "%.*s\n",status_buffer.length, 
                  status_buffer.value, 0);
            gss_release_buffer(&min_stat, &status_buffer);
        };
    } while (ctx != 0);

    if (GSS_ERROR(gss_display_status(&min_stat,
                                   min,
                                   GSS_C_MECH_CODE,
                                   GSS_C_NULL_OID,
                                   &ctx,
                                   &status_buffer))) {
      LDAPDebug(LDAP_DEBUG_ANY,"Error translating minor status code (%X)\n", 
              min, 0, 0);
    } else {
      LDAPDebug(LDAP_DEBUG_ANY, "%.*s\n", 
              status_buffer.length, status_buffer.value, 0);
      gss_release_buffer(&min_stat, &status_buffer);
    };
}

int
LDAP_CALL
ldap_gssapi_bind(
    LDAP        *ld,
    const char      *dn
)
{
    char *hostname;
    gss_buffer_desc send_tok, recv_tok, *token_ptr;
    struct berval send_ber, *recv_ber;
    OM_uint32 maj_stat, min_stat, min_ign;
    gss_name_t target_name;
    gss_ctx_id_t gss_context;
    int rc, conf_state;
    char sec_layer;
    int mms;
    unsigned char *up;
    int sd, sdi;
    
    hostname = ld->ld_defhost;
    if (hostname == NULL)
        return (LDAP_OTHER) ;

    LDAPDebug( LDAP_DEBUG_TRACE, "ldap_gss_bind_s to %s\n", hostname, 0, 0 );

    /* setup gss contexts */
    send_tok.value = malloc(strlen(hostname) + 5 + 1);
    if (send_tok.value == NULL) {
        return (LDAP_NO_MEMORY);
    }

    send_tok.length = strlen(hostname) + 5;
    strcpy(send_tok.value, "LDAP@");
    strcat(send_tok.value, hostname);

    LDAPDebug( LDAP_DEBUG_TRACE, "service name %s\n", send_tok.value, 0, 0 );

    maj_stat = gss_import_name(&min_stat, &send_tok,
                               gss_nt_service_name, &target_name) ;

    gss_release_buffer(&min_ign, &send_tok);
    if (GSS_ERROR(maj_stat)) {
        gss_display_error("gss_import_name", maj_stat, min_stat) ;
        return (LDAP_OTHER) ;
    }
    LDAPDebug(LDAP_DEBUG_TRACE, "Imported server name\n", 0, 0,0) ;

    token_ptr = GSS_C_NO_BUFFER;
    gss_context = GSS_C_NO_CONTEXT;

    do {
        maj_stat = 
            gss_init_sec_context(&min_stat,
                                 GSS_C_NO_CREDENTIAL,
                                 &gss_context,
                                 target_name,
                                 GSS_C_NULL_OID,
                                 GSS_C_MUTUAL_FLAG | GSS_C_INTEG_FLAG,
                                 60*60,
                                 GSS_C_NO_CHANNEL_BINDINGS,
                                 token_ptr,
                                 NULL,
                                 &send_tok,
                                 NULL,
                                 NULL) ;

        if (token_ptr != GSS_C_NO_BUFFER)
            gss_release_buffer(&min_ign, &recv_tok);
        if (maj_stat!=GSS_S_COMPLETE && maj_stat!=GSS_S_CONTINUE_NEEDED) {
            gss_display_error("gss_init_sec_context", maj_stat, min_stat) ;
            gss_release_name(&min_stat, &target_name);
            if (gss_context != GSS_C_NO_CONTEXT)
                gss_delete_sec_context(&min_stat, &gss_context,
                                       GSS_C_NO_BUFFER);
            return (LDAP_OTHER);
        }
        send_ber.bv_len = send_tok.length;
        send_ber.bv_val = (char *) send_tok.value;
        rc = ldap_sasl_bind_s (ld, dn, "GSSAPI", &send_ber,
                                   NULL, NULL, &recv_ber);
        gss_release_buffer(&min_ign, &send_tok);
        if (rc != LDAP_SUCCESS && rc != LDAP_SASL_BIND_IN_PROGRESS) {
            gss_release_name(&min_stat, &target_name);
            gss_delete_sec_context(&min_stat, &gss_context,
                                   GSS_C_NO_BUFFER);
            return (LDAP_OTHER);
        }
        recv_tok.length = recv_ber->bv_len;
        recv_tok.value = recv_ber->bv_val;

        if (maj_stat == GSS_S_CONTINUE_NEEDED) {
            token_ptr = &recv_tok;
        }
    } while (maj_stat == GSS_S_CONTINUE_NEEDED);

    gss_release_name(&min_stat, &target_name);

    maj_stat = gss_unwrap(&min_stat, gss_context, &recv_tok, &send_tok,
                          &conf_state, NULL);
    if (GSS_ERROR(maj_stat)) {
        gss_display_error("gss_unwrap", maj_stat, min_stat) ;
        return (LDAP_OTHER) ;
    }
    if (! conf_state) {
        LDAPDebug(LDAP_DEBUG_ANY,
                  "Message not encrypted %x\n",conf_state,0,0) ;
    }

    if (send_tok.length != 4)
        return(LDAP_DECODING_ERROR);

    up = (unsigned char *) send_tok.value;
    sec_layer = up[0];
    mms = (((up[1] << 8) + up[2]) << 8) + up[3];
    LDAPDebug(LDAP_DEBUG_ANY,
          "called unwrap: suppl info  %x sec offered %x mms %d\n",
          GSS_SUPPLEMENTARY_INFO(maj_stat), sec_layer, mms) ;

    recv_tok.length = 4;
    recv_tok.value = send_tok.value;
    *(char *)recv_tok.value = 6;

    maj_stat = gss_wrap(&min_stat, gss_context, 1, GSS_C_QOP_DEFAULT,
                        &recv_tok, &conf_state, &send_tok);
    if (GSS_ERROR(maj_stat)) {
        gss_display_error("gss_wrap", maj_stat, min_stat) ;
        return (LDAP_OTHER) ;
    }
    LDAPDebug(LDAP_DEBUG_ANY, "called wrap: suppl info  %x\n",
          GSS_SUPPLEMENTARY_INFO(maj_stat),0,0) ;

    send_ber.bv_len = send_tok.length;
    send_ber.bv_val = (char *) send_tok.value;
    rc = ldap_sasl_bind_s (ld, dn, "GSSAPI", &send_ber,
                           NULL, NULL, &recv_ber);
    recv_tok.length = recv_ber->bv_len;
    recv_tok.value = recv_ber->bv_val;

    if (rc != LDAP_SUCCESS) {
          (void) gss_release_buffer(&min_stat, &send_tok);
          return (LDAP_OTHER);
    }
    rc = ber_sockbuf_set_option( ld->ld_sbp,
            LBER_SOCKBUF_OPT_READ_FN, (void *) gss_read );
    rc |= ber_sockbuf_set_option( ld->ld_sbp,
            LBER_SOCKBUF_OPT_WRITE_FN, (void *) gss_write );
    rc |= ber_sockbuf_get_option( ld->ld_sbp,
            LBER_SOCKBUF_OPT_DESC, &sd );

    for (sdi=0; sdi<NSOCKD; sdi++)
        if (sockd_list[sdi].sd == sd)
            break;
    if (sdi >= NSOCKD) {
        for (sdi=0; sdi<NSOCKD; sdi++)
            if (sockd_list[sdi].sd == 0)
                break;
        if (sdi >= NSOCKD)
            return (LDAP_OTHER);
        sockd_list[sdi].sd = sd;
    }

    sockd_list[sdi].gss_context = gss_context;
    sockd_list[sdi].len = 0;
    return( LDAP_SUCCESS );
}

size_t
gss_read(int sd, char *buf, size_t len)
{
    int dlen; int i = 0;
    int sdi, rc, cnt;
    char *p;
    gss_buffer_desc gss_in ;
    gss_buffer_desc gss_out ;
    OM_uint32 maj_stat, min_stat ;
    int conf_state;

    for (sdi=0; sdi<NSOCKD; sdi++)
        if (sockd_list[sdi].sd == sd)
            break;
    if (sdi >= NSOCKD)
        return(-1);

    if (sockd_list[sdi].len <= 0) {
        for (i=0; i < 4; i+=rc) {
            rc = read( sd, &dlen+i, 4-i);
            if (rc <= 0)
                return(-1);
        }
        dlen = ntohl(dlen);
        p = malloc(dlen);
        if (p == NULL)
            return(-1);
        for (i=0; i < dlen; i+=rc) {
            rc = read(sd, p+i, dlen-i);
            if (rc <= 0)
                return(-1);
        }
        gss_in.length = dlen ;
        gss_in.value =  p;

        maj_stat = gss_unwrap(&min_stat, sockd_list[sdi].gss_context,
                              &gss_in, &gss_out, &conf_state, NULL);

        if(GSS_ERROR(maj_stat)) {
            gss_display_error("ber_fil unwrap", maj_stat, min_stat) ;
            return -1 ;
        }
        free(p);
        sockd_list[sdi].buf = sockd_list[sdi].p = gss_out.value;
        sockd_list[sdi].len = gss_out.length;
    }

    cnt = sockd_list[sdi].len > len ? len : sockd_list[sdi].len;
    memcpy(buf, sockd_list[sdi].p, cnt);
    sockd_list[sdi].len -= cnt;
    sockd_list[sdi].p += cnt;
    if (sockd_list[sdi].len == 0)
        free(sockd_list[sdi].buf);
    return(cnt);
}

size_t
gss_write(int sd, char *buf, size_t len)
{
    gss_buffer_desc gss_in ;
    gss_buffer_desc gss_out ;
    OM_uint32 maj_stat, min_stat ;
    int taglen, lenlen ;
    int sdi, cnt, rc;
    char *p;
    long atag  ;
    char *tag_buf ;

    for (sdi=0; sdi<NSOCKD; sdi++)
        if (sockd_list[sdi].sd == sd)
            break;
    if (sdi >= NSOCKD)
        return(-1);

    gss_in.length = len ;
    gss_in.value = buf ;

    /* we'd try for packet privacy, but some dce implementations don't
    ** ship with it by default (like digital's).  So, we'll stick
    ** with just integrity for now.
    */
    maj_stat = gss_wrap(&min_stat, sockd_list[sdi].gss_context,
                        1, GSS_C_QOP_DEFAULT, &gss_in, NULL, &gss_out);

    if (GSS_ERROR(maj_stat)) {
        gss_display_error("gss_seal", maj_stat, min_stat) ;
        return -1 ;
    }
    cnt = gss_out.length;
    cnt = htonl(cnt);
    rc = write(sd, &cnt, 4);
    cnt = gss_out.length;
    p = gss_out.value;
    while (cnt > 0) {
        rc = write(sd, p, cnt);
        if (rc < 0)
            return(-1);
        cnt -= rc;
        p += rc;
    }
    free(gss_out.value);
    return(len);
}

